/*
  undistort-opengl - Realtime pincushion distortion correction
  Copyright 2008, Scott A. Kuhl
  All rights reserved.

  This file is part of undistort-opengl.

  Undistort-opengl is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  Undistort-opengl is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
  License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with undistort-opengl.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <GL/glew.h>

#ifndef _WIN32
#include <string.h>
#endif

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#ifdef _WIN32
#include <windows.h>
#endif


#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


#define DEBUG 0

/* For HMD usage, change WINDOW_WIDTH and WINDOW_HEIGHT to the dimensions
 * of your screen.  If you don't need the application to fill your screen,
 * you can also set these dimensions to something smaller than your screen.
 */
#define WINDOW_WIDTH 1280
#define WINDOW_HEIGHT 1024


/* Horizontal and vertical field of view of the HMD (used in the first rendering pass) */
#define HMD_HFOV (float)41.7
#define HMD_VFOV (float)33.3
float aspectRatio;  /* This is automatically computed from the field of views */



/* These are the dimensions of the texture that we render to in the first
 * pass.  Since we are using the OpenGL extension for rectangular textures,
 * the dimensions don't have to be powers of two, and can be rectangular.
 * These dimensions should be equal to or smaller than the WINDOW_WIDTH and
 * WINDOW_HEIGHT variables since we prerender into the window's
 * framebuffer.*/
#define PRERENDER_WIDTH WINDOW_WIDTH
#define PRERENDER_HEIGHT WINDOW_HEIGHT
static GLubyte prerender[PRERENDER_WIDTH*PRERENDER_HEIGHT*3];

#define FRAGMENT_PROGRAM_FILENAME "undistort-opengl.glsl"


/* k1 is the parameter we adjust to correct for pincushion distortion. */
static GLfloat k1 = .1;
#define K1_STEP .01

/* texScale is used to scale the prerendered texture (vertically and
 * horizontally) to get it to fill the center vertical and center
 * horizontal lines on the screen. */
static GLfloat texScale[2] = { 1, 1 };


static GLboolean scaleToFit = GL_TRUE;  /* Scale to fit the screen */
static GLboolean showInfo = GL_TRUE;    /* Toggle the information display */
static GLboolean secondPass = GL_TRUE;  /* Apply pre-distortion */


/* The checkerboard texture is internally stored as a CHECKER_SIZE x
 * CHECKER_SIZE texture */
#define CHECKER_SIZE 1024
static GLubyte checkerboard[CHECKER_SIZE][CHECKER_SIZE][3];

/* References to our textures and fragment program */
static GLuint prerenderTexName, checkerboardTexName;
static GLhandleARB fragProgram;

/* Used to rotate the animated teapot */
static GLint teapot_angle = 0;


/* Frames per second variables */
static int frame=0,timebase=0;
static float fps=-1;




/* This points to the function which draws the scene (checkerboard or teapot). */
void drawCheckerboard();
void (*drawFunction)() = drawCheckerboard;

/* Check for OpenGL errors. */
#define isError() isErrorCheck(__FILE__, __LINE__)
int isErrorCheck(char *file, int line)
{
#if DEBUG
	GLenum code;
	if((code = glGetError()) != GL_NO_ERROR)
	{
		fprintf(stderr, "%s:%d OpenGL Error: %s\n", file, line, gluErrorString(code));
		return 1;
	}
	return 0;
#else
	return 0;
#endif
}



/* Reads in a text file and returns a char* which should be subsequently
 * free()'d */
char *textFileRead(char *filename)
{
	int chunkSize=1024;             /* read in chunkSize bytes at a time */
	int contentSpace = chunkSize;   /* space in 'content' array */

	/* We add one more character to create room to store a '\0' at the
	 * end. */
	char *content = malloc(sizeof(char)*(contentSpace+1));

	/* Pointer to where next chunk should be stored */
	char *contentLoc = content;

	FILE *fp = fopen(filename,"rt");
	int readChars;

	if( fp == NULL)
	{
		fprintf(stderr, "ERROR: Can't open %s\n", filename);
		exit(1);
	}

	do
	{
		readChars = fread(contentLoc, sizeof(char), chunkSize, fp);
		contentLoc[readChars] = '\0';
		contentSpace += chunkSize;
		content = realloc(content, sizeof(char)*(contentSpace+1));
		contentLoc = content + contentSpace - chunkSize;

	} while( readChars == chunkSize );

	/* We should now be at end of file. If not, there was an error. */
	if(feof(fp) == 0)
	{
		fprintf(stderr, "ERROR: Can't read %s\n", filename);
		exit(1);
	}

	fclose(fp);
	return content;
}


/* Gets the location of variable "name" in fragment/vertex program "program" */
GLint getUniLoc(GLhandleARB program, const GLchar *name)
{
	GLint loc = glGetUniformLocationARB(program, name);
	isError();

	if(loc == -1)
	{
		fprintf(stderr, "ERROR: No uniform variable in glsl program named %s\n", name);
#if DEBUG
		exit(1);
#endif
	}
	return loc;
}


/* Initialize a checkerboard texture */
void init_checkerboard()
{
	int i, j;
	static const int check_size = 64;
	for (i = 0; i < CHECKER_SIZE; i++)
	{
		for (j = 0; j < CHECKER_SIZE; j++)
		{
			if(((i/check_size % 2) + (j/check_size % 2)) == 1)
			{
				checkerboard[i][j][0] = 0;
				checkerboard[i][j][1] = 128;
				checkerboard[i][j][2] = 0;
			}
			else
			{
				checkerboard[i][j][0] = 0;
				checkerboard[i][j][1] = 0;
				checkerboard[i][j][2] = 0;
			}
		}
	}

	glGenTextures(1, &checkerboardTexName);
	glBindTexture(GL_TEXTURE_2D, checkerboardTexName);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glPixelStorei(GL_UNPACK_ALIGNMENT,1);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, CHECKER_SIZE, CHECKER_SIZE,
	             0, GL_RGB, GL_UNSIGNED_BYTE, checkerboard);
	glPixelStorei(GL_UNPACK_ALIGNMENT,4);
	isError();
}

/* Initialize our fragment program */
void init_shader()
{
	GLhandleARB fragShader = 0;
	char *fragText = NULL;
	char logString[1024];
	GLsizei actualLen = 0;
	GLint linked;

	if(!GLEW_ARB_fragment_program)
	{
		fprintf(stderr, "ERROR: fragment programs not supported\n");
		exit(1);
	}

	printf("Using fragment shader %s\n", FRAGMENT_PROGRAM_FILENAME);
	fragShader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER);
	fragText = textFileRead(FRAGMENT_PROGRAM_FILENAME);
	glShaderSource(fragShader, 1, (const char**) &fragText, NULL);
	free(fragText);

	/* Compile fragment program */
	glCompileShader(fragShader);
	glGetInfoLogARB(fragShader, 1024, &actualLen, logString);
	if(actualLen > 0)
		printf("Fragment shader log:\n%s\n", logString);


	fragProgram = glCreateProgram();
	if(fragProgram == 0)
	{
		fprintf(stderr, "ERROR: Failed to create fragment program.\n");
		exit(1);
	}
	glAttachShader(fragProgram, fragShader);

	isError();
	glLinkProgram(fragProgram);
	glUseProgram(fragProgram);
	isError();

	glValidateProgram(fragProgram);

	/* Check if program was glLinkProgram was successful. */
	glGetProgramiv((GLuint)fragProgram, GL_LINK_STATUS, &linked);
	isError();

	if(linked == GL_FALSE)
	{
		fprintf(stderr, "Failed to link fragment program.\n");
		exit(1);
	}

	glGetProgramInfoLog(fragProgram, 1024, &actualLen, logString);
	if(actualLen > 0)
		printf("Program log:\n%s\n", logString);

	glUseProgram(0);
}

/* Initialize the rectangular texture we are going to prerender into. */
void init_prerender()
{
	GLint textureSize = 0;
	int i;

	if(!GL_ARB_texture_rectangle)
	{
		fprintf(stderr, "ERROR: rectangular textures not supported\n");
		exit(1);
	}

	glGetIntegerv(GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB, &textureSize);
	if(textureSize < PRERENDER_WIDTH ||
	   textureSize < PRERENDER_HEIGHT)
	{
		fprintf(stderr, "ERROR: You requested a texture size of %dx%d but only %dx%d is supported by rectangular textures\n",
		        PRERENDER_WIDTH, PRERENDER_HEIGHT, (int)textureSize, (int)textureSize);
		exit(1);
	}

	glActiveTexture(GL_TEXTURE0);
	glGenTextures(1, &prerenderTexName);
	glBindTexture(GL_TEXTURE_RECTANGLE_ARB, prerenderTexName);

	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	for(i=0; i<PRERENDER_WIDTH*PRERENDER_HEIGHT*3; i++)
		prerender[i] = 1;

	glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB,
	             PRERENDER_WIDTH, PRERENDER_HEIGHT, 0,
	             GL_RGB, GL_UNSIGNED_BYTE, prerender);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	glDisable(GL_TEXTURE_RECTANGLE_ARB);

}

/* Initialize a lighting/materials for the rotating teapot */
void init_light()
{

	GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
	GLfloat mat_shininess[] = { 50.0 };
	GLfloat light_position[] = { 5.0, 5.0, 0.0, 0.0 };
	GLfloat ambient[] = { .7, .7, .7, 1 };

	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

}

void init_multisample()
{
	if(GL_ARB_multisample)
	{
		GLint bufs, samples;
		glGetIntegerv(GL_SAMPLE_BUFFERS_ARB, &bufs);
		glGetIntegerv(GL_SAMPLES_ARB, &samples);
		printf("\n");
		printf("Multisampling info: buffers=%d samples=%d\n", (int)bufs, (int)samples);

		if( bufs == 0 || samples <= 1)
		{
			printf("WARNING: Your video card indicates that it supports multisampling,\n");
			printf("  but we failed to turn it on.\n\n");
			printf("  You may have a broken version of glut.\n");
			printf("  Note that some versions of freeglut are broken unless patched.\n");
			printf("  See: http://www.nabble.com/GLUT_MULTISAMPLE-t1649080.html\n\n");
			return;
		}
		else
			glEnable(GL_MULTISAMPLE_ARB);
	}
	else
	{
		printf("Warning: Multisampling not supported by your video card.  Running anyway...\n");
	}

}



void drawCheckerboard()
{
	glDisable(GL_LIGHTING);
	glViewport(0,0,PRERENDER_WIDTH, PRERENDER_HEIGHT);

	/* Clear the buffer */
	glClearColor(0,1,0,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* Setup the camera */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1, 1, -1, 1, .1, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -10);

	/* Turn on texturing */
	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
	glBindTexture(GL_TEXTURE_2D, checkerboardTexName);

	/* Draw the textured quad */
	glBegin(GL_QUADS);
	glTexCoord2f(1,0); glVertex3f( 1,-1, 0);
	glTexCoord2f(1,1); glVertex3f( 1, 1, 0);
	glTexCoord2f(0,1); glVertex3f(-1, 1, 0);
	glTexCoord2f(0,0); glVertex3f(-1,-1, 0);
	glEnd();

	/* Restore state */
	glFlush();
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
}

/* Draw a scene with dimensions PRERENDER_WIDTH x PRERENDER_HEIGHT */
void drawTeapot()
{
	glEnable(GL_LIGHTING);
	glViewport(0, 0, PRERENDER_WIDTH, PRERENDER_HEIGHT);

	/* Clear the buffer */
	glClearColor(.2,.2,.2,1);  /* dark gray */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* Set up camera */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();


	gluPerspective(HMD_HFOV, aspectRatio, 1.0, 30.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	/* Draw the scene */
	glTranslatef(0, 0, -10);
	glRotatef(teapot_angle, 0, 1, 0);

	glFrontFace(GL_CW);
	glutSolidTeapot(2);
	glFrontFace(GL_CCW);

	glDisable(GL_LIGHTING);
}

/* Draws the second pass---filling up WINDOW_WIDTH x WINDOW_HEIGHT */
void drawSecondPass()
{
	/* start second pass */
	glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

	/* For debugging: Clear the screen---we shouldn't need to do this
	 * because the textured quad we put in front of the camera should fill
	 * the screen. */
#if DEBUG
	/* Fill the screen with red */
	glClearColor(1,0,0,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	/* If you still see red after the second pass, the second pass didn't
	 * fill the entire screen */
#else
	glDisable(GL_DEPTH_TEST);  /* No need for depth testing */
#endif

	glDisable(GL_LIGHTING);

	/* Setup the camera */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1, 1, -1, 1, .1, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -10);

	/* Turn on texturing */
	glEnable(GL_TEXTURE_RECTANGLE_ARB);

	/* Turn on fragment program */
	glUseProgramObjectARB(fragProgram);
	glUniform1fARB(getUniLoc(fragProgram, "k1")          , k1);
	glUniform2fARB(getUniLoc(fragProgram, "texScale")    , texScale[0], texScale[1]);
	glUniform2fARB(getUniLoc(fragProgram, "screenDimen") , PRERENDER_WIDTH, PRERENDER_HEIGHT);
	glUniform1iARB(getUniLoc(fragProgram, "tex"), 0 );

	/* Draw the textured quad */
	glBegin(GL_QUADS);
	glTexCoord2f(PRERENDER_WIDTH,0);                glVertex3f( 1,-1, 0);
	glTexCoord2f(PRERENDER_WIDTH,PRERENDER_HEIGHT); glVertex3f( 1, 1, 0);
	glTexCoord2f(0,              PRERENDER_HEIGHT); glVertex3f(-1, 1, 0);
	glTexCoord2f(0,              0);                glVertex3f(-1,-1, 0);
	glEnd();

	/* Restore the state... */
	glUseProgramObjectARB(0);            /* Turn off fragment program */
	glDisable(GL_TEXTURE_RECTANGLE_ARB); /* Turn off texturing */
	glEnable(GL_DEPTH_TEST);             /* Turn on depth test... */
	glEnable(GL_LIGHTING);
}

void drawInfo()
{
	char text[1024];
	int i;

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glColor3f(0,.8,0);

	/* Setup the camera */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1, 1, -1, 1, .1, 100);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -10);

	glRasterPos2f(-.98, .8);


	sprintf(text, "correction is %s;  k1=%0.3f;   texScale.x=%0.3f;   texScale.y=%0.3f;   fps=%0.1f\n",
	        secondPass==GL_TRUE ? "on" : "off", k1, texScale[0], texScale[1], fps);

	for(i=0; i<strlen(text); i++)
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, text[i]);

	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
}

void display(void)
{

	/* Draw the scene in the first pass */
	drawFunction();

	/* Check for errors */
	isError();

	if(secondPass)
	{
		/* Copy the scene on framebuffer into a texture */
		glBindTexture(GL_TEXTURE_RECTANGLE_ARB, prerenderTexName);
		glCopyTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB, 0, 0,
		                 PRERENDER_WIDTH, PRERENDER_HEIGHT, 0);

		/* Draw the second pass */
		drawSecondPass();

		isError();
	}

	if(showInfo)
		drawInfo();

	/* Draw the result on the screen */
	glutSwapBuffers();
}

static GLboolean warnedAboutSmallWindow = GL_FALSE;
void reshape(int w, int h)
{
	if(warnedAboutSmallWindow && (PRERENDER_WIDTH  > w || PRERENDER_HEIGHT > h))
	{
		fprintf(stderr, "WARNING: Window size must be at least %d x %d---i.e. at least as large as the prerender texture.\n",
		        PRERENDER_WIDTH, PRERENDER_HEIGHT);
		warnedAboutSmallWindow = GL_TRUE;
	}
}


void idle()
{
	int time;

	teapot_angle++;
	if(teapot_angle >= 360)
		teapot_angle = 0;

	/* Calculate frames per second */
	frame++;
	time=glutGet(GLUT_ELAPSED_TIME);
	if (time - timebase > 1000)
	{
		fps = frame*1000.0/(time-timebase);
		timebase = time;
		frame = 0;
	}

	glutPostRedisplay();
}


void change_k1(GLfloat newk1)
{
	float horizDist;
	GLfloat norm = 0;
	GLfloat vertDist = 0;

	k1 = newk1;

	if(!scaleToFit)
	{
		texScale[0] = texScale[1] = 1;
		return;
	}

	/* In the fragment program, we normalize values to the distance from
	 * the center of the image to the corner. */
	norm = sqrt((PRERENDER_HEIGHT/2.0)*(PRERENDER_HEIGHT/2.0) +
	            (PRERENDER_WIDTH/2.0) *(PRERENDER_WIDTH/2.0));

	/* Normalized distance from center to top of image. */
	vertDist = (PRERENDER_HEIGHT / 2.0) / norm;

	/* Determine how the top center pixel maps onto the prerendered
	 * texture.  Then, compute how much we need to scale our texture
	 * lookups to get the final image to fit the screen vertically. */
	texScale[1] = vertDist / (vertDist + k1*vertDist*vertDist*vertDist);

	/* Do the same for the horizontal */
	horizDist = (PRERENDER_WIDTH / 2.0) / norm;
	texScale[0] = (horizDist / ((horizDist + k1*horizDist*horizDist*horizDist)));
}



void keyboard(unsigned char key, int x, int y)
{
	switch(key)
	{
		case 27:
		case 'q':
		case 'Q':
			exit(0);
			break;
		case 'f':
			glutFullScreen();
			break;
		case ' ':
			if(drawFunction == drawTeapot)
				drawFunction = drawCheckerboard;
			else
				drawFunction = drawTeapot;
			break;
		case 's':
			secondPass = !secondPass;
			break;
		case 'z':
			scaleToFit = !scaleToFit;
			change_k1( k1 );
			break;
		case 'm':
			change_k1( k1+K1_STEP );
			break;
		case 'n':
			change_k1( k1-K1_STEP );
			break;
		case 'i':
			showInfo = !showInfo;
			break;
	}

	glutPostRedisplay();
}



void init(void)
{
	printf("undistort-opengl %s - Realtime pincushion distortion correction\n", VERSION);
	printf("Copyright 2008 Scott A. Kuhl <percept-software@cs.utah.edu>\n");
	printf("This is free software; see the source for copying conditions.  There is NO\n");
	printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\n");

	aspectRatio = tan( HMD_HFOV*M_PI/180.0 / 2.0 ) / tan( HMD_VFOV*M_PI/180.0 / 2.0 );
	printf("HFOV=%0.3f, VFOV=%0.3f, aspectRatio (as computed by this program)=%0.3f\n", HMD_HFOV, HMD_VFOV, aspectRatio);

	/* Initialize variables that change with k1 */
	change_k1(k1);

	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_SMOOTH);

#if DEBUG
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
#endif

	init_light();
	init_checkerboard();
	init_multisample();
	init_prerender();
	init_shader();

	printf("\nKeyboard controls:\n");
	printf("f - enter full screen mode.\n");
	printf("i - toggle onscreen information\n");
	printf("z - toggle scaling to fit screen\n");
	printf("s - toggle distortion correction\n");
	printf("space - toggle between checkerboard and teapot\n");
	printf("m and n - change k1 parameter for distortion correction\n");
	printf("ESC - exit\n");

	printf("This program may not function correctly if application window is clipped by the edge of your screen.\n");
}


int main(int argc, char** argv)
{
	char windowTitle[1024];

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_MULTISAMPLE);
	glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
	glutInitWindowPosition(0,0);
	sprintf(windowTitle, "undistort-opengl %s", VERSION);
	glutCreateWindow(windowTitle);

	int glew_err = glewInit();
	if(glew_err != GLEW_OK)
		fprintf(stderr, "Error: %s\n", glewGetErrorString(glew_err));

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutIdleFunc(idle);

	/* Try to drop into full screen mode.  May not work with some versions
	 * of freeglut. */
	glutFullScreen();

	init();

	glutMainLoop();
	return 0;
}
